#ifndef SMPBASICBLOCK_H
#define SMPBASICBLOCK_H 1

// SMPBasicBlock.h
//
// This header defines the interfaces needed for analyzing basic blocks.

#include <list>
#include <set>

#include <cstddef>

#include <pro.h>
#include <ida.hpp>
#include <ua.hpp>

#include "SMPDataFlowAnalysis.h"
#include "SMPInstr.h"

using namespace std;

// Do Full Liveness Analysis?  Experimental by KBinswanger
#define SMP_FULL_LIVENESS_ANALYSIS 1

// Value for a block number before it is initialized.
#define SMP_BLOCKNUM_UNINIT (-1)

// Value for an SSA subscript number before it is initialized by SSA renaming.
#define SMP_SSA_UNINIT  (-1)

// Class defining basic blocks.
class SMPBasicBlock {
public:
	// Constructors
	SMPBasicBlock(SMPFunction *Func, list<SMPInstr>::iterator FirstInstr, list<SMPInstr>::iterator LastInstr);
	// Operators
	int operator==(const SMPBasicBlock &rhs) const;
	// Get methods
	ea_t GetFirstAddr(void) const;
	inline int GetNumber(void) const { return BlockNum; };
	inline SMPBasicBlock *GetThisBlock(void) { return this; };
	inline SMPFunction *GetFunc(void) { return MyFunc; };
	inline list<list<SMPInstr>::iterator>::iterator GetFirstInstr(void) { return Instrs.begin(); };
	inline list<list<SMPInstr>::iterator>::iterator GetLastInstr(void) { return Instrs.end(); };
	inline ea_t GetFirstFlagsDef(void) const { return FirstFlagsDef; };
	inline ea_t GetLastFlagsUse(void) const { return LastFlagsUse; };
	inline list<list<SMPBasicBlock>::iterator>::iterator GetFirstPred(void) {
		return Predecessors.begin();
	};
	inline list<list<SMPBasicBlock>::iterator>::iterator GetLastPred(void) {
		return Predecessors.end();
	};
	inline list<list<SMPBasicBlock>::iterator>::iterator GetFirstSucc(void) {
		return Successors.begin();
	};
	inline list<list<SMPBasicBlock>::iterator>::iterator GetLastSucc(void) {
		return Successors.end();
	};
	inline size_t GetNumPreds(void) const { return Predecessors.size(); };
	set<op_t, LessOp>::iterator GetFirstLiveIn(void); // LiveInSet.begin()
	set<op_t, LessOp>::iterator GetLastLiveIn(void); // LiveInSet.end()
	set<op_t, LessOp>::iterator GetFirstLiveOut(void); // LiveOutSet.begin()
	set<op_t, LessOp>::iterator GetLastLiveOut(void); // LiveOutSet.end()
	set<op_t, LessOp>::iterator GetFirstVarKill(void); // KillSet.begin()
	set<op_t, LessOp>::iterator GetLastVarKill(void); // KillSet.end()
	set<op_t, LessOp>::iterator GetFirstUpExposed(void); // UpExposedSet.begin()
	set<op_t, LessOp>::iterator GetLastUpExposed(void); // UpExposedSet.end()
	set<op_t, LessOp>::iterator GetFirstLocalName(void); // LocalNames.begin()
	set<op_t, LessOp>::iterator GetLastLocalName(void); // LocalNames.end()
	set<int>::iterator GetFirstDomFrontier(void); // DomFrontier.begin()
	set<int>::iterator GetLastDomFrontier(void); // DomFrontier.end()
	set<SMPPhiFunction, LessPhi>::iterator GetFirstPhi(void); // PhiFunctions.begin()
	set<SMPPhiFunction, LessPhi>::iterator GetLastPhi(void); // PhiFunctions.end()
	int GetPredPosition(int BlockNum); // What is position # within Preds of BlockNum?
	set<SMPPhiFunction, LessPhi>::iterator FindPhi(op_t FindOp);
	set<op_t, LessOp>::iterator FindLocalName(op_t FindOp);
	// Set methods
	inline void SetProcessed(bool flag) { Processed = flag; };
	inline void SetShared(void) { SharedTailChunk = true; };
	inline void SetNumber(int Num) { BlockNum = Num; };
	void LinkToPred(list<SMPBasicBlock>::iterator Predecessor);
	void LinkToSucc(list<SMPBasicBlock>::iterator Successor);
	inline void AddVarKill(op_t KillOp) { KillSet.insert(KillOp); return; };
	bool ErasePhi(op_t OldOp); // erase() from phi function list
	void ErasePred(ea_t FirstAddr); // erase() block starting at FirstAddr from Preds list
	bool AddPhi(SMPPhiFunction NewPhi);
	bool SetPhiDefSSA(op_t DefOp, int SSANum);
	bool SetPhiUseSSA(op_t UseOp, size_t index, int SSANum);
	set<SMPPhiFunction, LessPhi>::iterator SetPhiDefType(op_t DefOp, SMPOperandType Type);
	bool SetPhiUseType(op_t DefOp, size_t index, SMPOperandType Type);
	set<SMPPhiFunction, LessPhi>::iterator SetPhiDefMetadata(op_t DefOp, SMPMetadataType Status);
	// Query methods
	inline bool IsProcessed(void) const { return Processed; };
	inline bool HasIndirectJump(void) const { return IndirectJump; };
	inline bool HasReturn(void) const { return Returns; };
	inline bool IsShared(void) const { return SharedTailChunk; };
	bool AllNops(void); // Are all instructions in the block no-ops?
	inline bool IsLiveIn(op_t CurrOp) const {
		return (LiveInSet.end() != LiveInSet.find(CurrOp));
	}
	inline bool IsLiveOut(op_t CurrOp) const {
		return (LiveOutSet.end() != LiveOutSet.find(CurrOp));
	}
	inline bool IsVarKill(op_t CurrOp) const {
		return (KillSet.end() != KillSet.find(CurrOp));
	}
	bool IsLocalName(op_t CurrOp) const;
	// Printing methods
	void Dump(void);
	// Analysis methods
	bool AllPredecessorsNumbered(void);
	void Analyze();
	void InitKilledExposed(void); // Initialize KilledSet and UpExposedSet
	bool UpdateLiveOut(void); // Iterate once on updating LiveOutSet; return true if changed
	void AddToDomFrontier(int); // Add RPO block number to DomFrontier set.
	void SetLocalNames(void); // Fille the LocalNames member set
	void SSALocalRenumber(void); // Renumber references to local names
#if SMP_FULL_LIVENESS_ANALYSIS
	void CreateGlobalChains(void); // Create DEF-USE chains for global names used here
	bool IsGlobalRegDead(ea_t InstAddr, op_t Operand, unsigned int RegIndex) const; // Is global reg dead at InstAddr?
#endif
	bool IsRegDead(ea_t InstAddr, unsigned int RegIndex) const; // Is local reg dead at InstAddr?
	void MarkDeadRegs(void); // Find dead registers for each mmStrata-instrumented instruction
	bool PropagateLocalDefType(op_t DefOp, SMPOperandType DefType, ea_t DefAddr, int SSANum); // to all uses
	bool InferLocalDefType(op_t DefOp, unsigned int LocIndex, ea_t DefAddr); // Can DEF type be inferred from all USEs?
	bool PropagateGlobalDefType(op_t DefOp, SMPOperandType DefType, int SSANum); // to all uses
	bool InferAllPhiDefTypes(void); // infer, propagate to all uses
	bool PropagateLocalMetadata(op_t UseOp, SMPMetadataType Status, int SSANum);
	bool FindRedundantLocalMetadata(bool SafeFunc); // consecutive DEFs, same type => redundant
private:
	// Data
	ea_t FirstAddr;
	int BlockNum;  // Number for block ordering algorithms
	SMPFunction *MyFunc; // function containing this block
	bool Processed; // Set to true to prevent infinite recursions in CFG traversals

		// cached query results
	bool IndirectJump; // contains an indirect jump instruction
	bool Returns;      // contains a return instruction
	bool SharedTailChunk; // is part of a code chunk shared among functions
	bool IsLiveInStale; // Has LiveOutSet changed since LiveInSet was computed?
	bool FlagsDeadAfterLastUse; // flags global, dead after last use in the block
	bool FlagsDeadBeforeFirstKill; // flags global, dead before first DEF in block
	ea_t FirstFlagsDef; // addr of first instr that DEFs the flags
	ea_t LastFlagsUse;  // addr of last instr that USEs the flags

	list<list<SMPInstr>::iterator> Instrs;
	list<list<SMPBasicBlock>::iterator> Predecessors;
	list<list<SMPBasicBlock>::iterator> Successors;
		// Four sets used in live variable analysis
	set<op_t, LessOp> KillSet;   // variables killed in this block
	set<op_t, LessOp> UpExposedSet; // upward exposed variables in this block
	set<op_t, LessOp> LiveOutSet; // Live-Out variables in this block
	set<op_t, LessOp> LiveInSet; // contribution to predecessor's live-out iteration
		// SSA data structures
	set<int> DomFrontier; // Dominance frontier for block, as set of RPO block numbers
	set<SMPPhiFunction, LessPhi> PhiFunctions; // SSA incoming edge phi functions
	set<op_t, LessOp> LocalNames; // non-global names referenced in this block
	SMPCompleteDUChains LocalDUChains; // def-use chains for local names
#if SMP_FULL_LIVENESS_ANALYSIS
	SMPCompleteDUChains GlobalDUChains; // def-use chains for global names
#endif

	// Methods
	bool MDAlreadyKilled(op_t) const; // Was op_t killed by something already in KillSet?
	bool AreGlobalFlagsDead(ea_t, op_t FlagOp); // Is the flags register FlagOp dead at ea_t?
	set<SMPPhiFunction, LessPhi>::iterator InferPhiDefType(set<SMPPhiFunction, LessPhi>::iterator DefPhi, bool &changed); // infer, propagate to all uses
}; 

#endif
